{$A-}  {Rekursion erlaubt}
Program Difmathebaum;

{Das Programm wandelt eine eingegebene Funktion in einen Binaerbaum um,}
{differenziert sie und gibt den Ergebnisbaun wieder aus}

Const   Version = '[2.3]';

Type    Ptr     = ^Element;
        Strings = String[5];          {Knoteninhalt}
        Stringn = String[100];        {Funktion in mathematischer Schreibweise}
        Element = Record
                    Wert  : Strings;
                    l, R  : Ptr       {Zeiger auf dem linken und rechten Ast}
                  End;

Var     Funktion : Stringn;
        Variable : Strings;
        Antwort  : Char;
        Stack    : ^Integer;          {Fuer Stack-Verwaltung}

{--------------------  Utility-Funktionen  ------------------------}

Function Upstring (F:Stringn):Stringn;
{Wandelt Kleinbuchstaben in F in Grossbuchstaben um}
Var     I : Integer;
Begin
  For I := 1 to Length(F) do  F[I] := Upcase(F[I]);
  Upstring := F
End;

Function Killblanks (F:Stringn):Stringn;
{Entfernt Leerzeichen aus F}
Var     I, N  : Integer;
Begin
  N := 1;
  For I := 1 to Length(F) do
    If F[I] <> ' ' then  Begin F[N] := F[I]; N := N+1 End;
  Killblanks := Copy(F,1,N-1)
End;

Function Search (F:Stringn):Integer;
{Sucht Rechenzeichen mit niedrigster Prioritaet und gibt seine Position}
{zurueck. Klammerausdruecke werden ueberlesen. Wenn kein Rechenzeichen da}
{ist, wird 0 uebergeben}
Var     I, Kl, P, Art  : Integer;
Begin
  P := 0; Kl := P; Art := 6;
  For I := Length(F) downto 1 do
  Begin
    Case F[I] of
      ')' : Kl := Kl+1;
      '(' : Kl := Kl-1
    End;
    If Kl = 0 then
      Case F[I] of
        '+' : If Art > 1 then Begin P := I; Art := 1 End;
        '-' : If Art > 2 then Begin P := I; Art := 2 End;
        '*' : If Art > 3 then Begin P := I; Art := 3 End;
        '/' : If Art > 4 then Begin P := I; Art := 4 End;
        '^' : If Art > 5 then Begin P := I; Art := 5 End
      End
  End;
  Search := P
End;

{-------------------- Funktionen zur Baumgenerierung ----------------------}

Function Makenew (Baum:Ptr):Ptr;
{Erzeugt einen neuen Knoten und besetzt die Aeste mit NIL}
Begin
  New(Baum);
  Mark(Baum);
  Baum^.L := Nil; Baum^.R := Nil; Makenew := Baum;
  Release(Baum)
End;

Function Konst (C:Strings):Ptr;
{Erzeugt einen neuen Knoten, besetzt die Aeste mit Nil und den Knoten}
{selbst mit dem Argument C}
Var     Baum  : Ptr;
Begin
  New(Baum);
  Mark(Baum);
  Baum^.Wert := C; Baum^.L := Nil; Baum^.R := Nil; Konst := Baum;
  Release(Baum)
End;

Function Add (Sign:Strings; U,V:Ptr):Ptr;
{Erzeugt einen neuen Knoten, der den Inhalt 'Sign' hat und dessen Aeste}
{auf die Baeume U und V zeigen}
Var     Baum  : Ptr;
Begin
  New(Baum);
  Mark(Baum);
  Baum^.Wert := Sign; Baum^.L := U; Baum^.R := V; Add := Baum;
  Release(Baum)
End;

{--------------------------- Hauptfunktionen --------------------------------}

Function Convert (F:Stringn):Ptr;
{Erzeugt einen Binaerbaum, der der eingegebenen Funktion entspricht}
Label   Nochmal;
Var     K, P  : Integer;
Begin
  Nochmal: K := Pos('(',F); P := Search(F); {Pos. des Rechenzeichens errechnen}
  If (P = 0) and (F[1] = '(') and (F[Length(F)] = ')') then
    Begin F := Copy(F,2,Length(F)-2); Goto Nochmal End; {Klammern wegnehmen}
  If (K = 0) and (P = 0) then Convert := Konst(F)
  Else {es muss rekursiv vorgegangen werden}
    If P = 0 then {Funktion z.B. sin(x)}
      Convert := Add(Copy(F,1,K-1),Convert(Copy(F,K+1,Length(F)-K-1)),Nil)
    Else {normaler Ausdruck}
      Convert := Add(Concat(F[P],''),Convert(Copy(F,1,P-1)),Convert(Copy(F,P+1,Length(F)-P)))
End;

Procedure Output(Baum:Ptr);
{Gibt den Baum im normalen mathematischen Format aus}
Var     S     : Strings;
        K     : Char;
        U, V  : Ptr;

  Procedure Lklamm(S,K:Strings; U:Ptr);
  {Erzeugt Klammern fuer linken Ast}
  Begin
    Case U^.Wert[1] of  {Entsprechend der Prioritaet u. der Stell. der Rechzeich. Klamm. setz}
      '+','-' : If (S <> '+') and (S <> '-') then Write(K);
      '*','/' : If S = '^' then Write(K)
    End
  End;

  Procedure Rklamm(S,K:Strings; V:Ptr);
  {Erzeugt Klammern fuer rechten Ast}
  Begin
    Case V^.Wert[1] of  {Entsprechend der Prioritaet u. der Stell. der Rechzeich. Klamm. setz}
      '+','-' : If S <> '+' then Write(K);
      '*','/' : If (S = '/') or (S = '^') then Write(K);
      '^'     : If S = '^' then Write(K)
    End
  End;

  Function Blatt (Baum:Ptr):Strings;
  {Liefert den Inhalt eines Knotens, falls er ein Blatt ist, sonst ''}
  Begin
    Blatt := '';
    If (Baum^.L = Nil) and (Baum^.R = Nil) then Blatt := Baum^.Wert
  End;

Begin
  New(U); Mark(U); New(V);
  S := Baum^.Wert; U := Baum^.L; V := Baum^.R;
  If (U = Nil) and (V = Nil) then Write(S)  {Wenn ein Blatt, dann den Inhalt ausgeben}
  Else
    If V = Nil then  {Funktion}
    Begin
      Write(S,'('); Output(U); Write(')')  {Rekursiv, nur linken Ast beruecksichtigen}
    End
    Else  {Normal}
    If (Blatt(U) = '0') and (Blatt(V) <> '') and (S = '-') then Write(S,V^.Wert) {0-Blatt}
    Else
    Begin
      Lklamm(S,'(',U); Output(U); Lklamm(S,')',U);  {Zuerst linken Ast rekursiv ausgeben}
      Write(S);  {Dann Rechenzeichen}
      Rklamm(S,'(',V); Output(V); Rklamm(S,')',V)   {Schliesslich den rechten Ast}
    End;
  Release(U)  {Stack freigeben}
End;

Function Dif (Baum:Ptr):Ptr; Forward;  {Weil sie schon in Prod benutzt wird}

Function Prod (Sign:Strings; U,V:Ptr):Ptr;
{Erzeugt einen Baum, der dem folgenden Ausdruck entspricht:}
{U'*V (+-) U*V', siehe Produkt- und Quotientenregel}
Begin
  Prod := Add(Sign,Add('*',Dif(U),V),Add('*',U,Dif(V)))
End;

Function Dif;
{Differenziert den Baum, ohne ihn zu vereinfachen}
Var     U, V, H1  : Ptr;
        S         : Strings;
Begin
  New(U); New(V);
  U := Baum^.L; V := Baum^.R; S := Baum^.Wert;  {Pointer besetzen}
  If (U = Nil) and (V = Nil) then  {Blatt: Konstante oder Variable}
    If S = Variable then Dif := Konst('1') Else Dif := Konst('0')
  Else
    If V = Nil then  {Funktion}
    Begin
      Dif := Konst('Error');  {Falls keine Funktion gefunden wird}
      If S = 'EXP' then Dif := Add('*',Add('EXP',U,Nil),Dif(U));       {EXP(X)'=EXP(X)}
      If S = 'LN'  then Dif := Add('*',Add('/',Konst('1'),U),Dif(U));  {LN(X)'=1/X}
      If S = 'SIN' then Dif := Add('*',Add('COS',U,Nil),Dif(U));       {SIN(X)'=COS(X)}
      If S = 'COS' then Dif := Add('*',Add('-',Konst('0'),Add('SIN',U,Nil)),Dif(U));
                                                                       {COS(X)'=-SIN(X)}
      If S = 'TAN' then Dif := Add('*',Add('+',Konst('1'),
                               Add('^',Add('TAN',U,Nil),Konst('2'))),Dif(U));
                                                                       {TAN(X)'=1+TAN(X)^2}
      If S = 'SQR' then Dif := Add('*',Add('/',Konst('1'),
                               Add('*',Konst('2'),Add('SQR',U,Nil))),Dif(U));
                                                                       {SQR(X)'=1/(2*SQR(X))}
    End  {Beliebig nach Schema F erweiterbar}
    Else {Normal}
      If Length(S) > 1 then Dif := Konst('Error')  {Rechenzeichen nicht bekannt, da zu lang}
      Else
        Case S[1] of
          '+','-' : Dif := Add(S,Dif(U),Dif(V));  {F=U+V, F'=U'+V'; F=U-V, F'=U'-V' SumReg.}
          '*'     : Dif := Prod('+',U,V);         {F=U*V, F'=U'*V+U*V' ProdReg.}
          '/'     : Dif := Add('/',Prod('-',U,V),Add('^',V,Konst('2')));
                                                  {F=U/V, F'=(U'*V-U*V')/V^2 QuotReg.}
          '^'     : Begin  {KettReg.}
                      New(H1);
                      H1 := Dif(V); If H1^.Wert = '0' then  {Potenz ist konstant}
                        {F=U^C, F'=C*U'*U^(C-1)}
                        Dif := Add('*',V,Add('*',Dif(U),Add('^',U,Add('-',V,Konst('1')))))
                      Else {Potenz ist Funktion}
                        {F=U^V, F'=U^V*(U'/U*V+LN(U)*V')}
                        Dif := Add('*',Add('^',U,V),Add('+',Add('*',Add('/',Dif(U),U),V),
                                       Add('*',Add('LN',U,Nil),H1)))
                    End
        End
End;

{--------------------------- Hauptprogramm -----------------------------}

Begin
  ClrScr;
  Write('Eingabe Differentiationsvariable: '); Readln(Variable);
  Variable := Upstring(Killblanks(Variable));
  Repeat
    ClrScr;
    Writeln('*** Differentiationsprogramm Version: ',Version,' ***');
    Writeln;
    Writeln('Eingabe F(',Variable,'), Ausgabe F''(',Variable,')');
    Writeln;
    Write('F (',Variable,') = ');
    Readln(Funktion);
    Funktion := Upstring(Killblanks(Funktion));
    Write('F''(',Variable,') = ');
    Mark(Stack);
    Output(Dif(Convert(Funktion)));  {Hauptaufruf}
    Writeln; Writeln;
    Release(Stack);
    Write('Noch eine Funktion (J/N) ? ');
    Repeat
      Read(KBD,Antwort);
    Until Upcase(Antwort) in ['J','N'];
  Until Upcase(Antwort) = 'N';
End.
