program LogiSim;
const
  Version       = '1.2';
  High          = true;
  Low           = false;
  MemSize       = 5000;
  MaxNodes      = 500;
  MaxView       = 30;
  Delimiter     = ' ';
  Invers        = #$1B#$29;
  Norm          = #$1B#$28;
  NumOfElements = 12;
type
  TString       = string[8];
  BigString     = string[96];
  SingleElement = record
    Name      :TString;
    Source    :boolean;
    MinNodes  :integer;
    ReqNodes  :integer;
    ReqMemory :integer
  end;
  NodeStates    = array[1..MaxNodes] of boolean;
  MemType       = array[0..15] of boolean;
var
  Mem         :array[1..MemSize] of integer;
  Pc          :integer;
  Node        :array[1..MaxNodes] of TString;
  NumOfNodes  :integer;
  ActNum      :integer;
  ActState    :NodeStates;
  RefState    :NodeStates;
  View        :array[1..MaxView] of integer;
  ViewPtr     :integer;
  Element     :array[1..NumOfElements] of SingleElement;
  ElementNo   :integer;
  Map         :text;
  MapName     :string[14];
  Line        :string[255];
  Piece       :array[1..80] of TString;
  NumOfPieces :integer;
  NumOfParts  :integer;
  Error       :boolean;
  Changes     :boolean;
  NewRun      :boolean;
  InStable    :TString;
  Q           :integer;

procedure PrintI(X,Y :integer; S :BigString);
begin
  gotoxy(X,Y);
  write(Invers,S)
end;

procedure PrintN(X,Y :integer; S :BigString);
begin
  gotoxy(X,Y);
  write(Norm,S)
end;

procedure ClrLine(I :integer);
begin
  PrintN(1,I,'                                                                                                ')
end;

function UsedMem(Op :integer):integer;
begin
  with Element[Op] do
    UsedMem:=1+ReqNodes+ReqMemory
end;

procedure Pack(M :MemType; var I :integer);
var
  Q :integer;
begin
  I:=0;
  for Q:=15 downto 0 do
  begin
    I:=I shl 1;
    I:=I or ord(M[Q])
  end
end;

procedure UnPack(I :integer; var M :MemType);
var
  Q :integer;
begin
  for Q:=0 to 15 do
  begin
    M[Q]:=I and 1 <> 0;
    I:=I shr 1
  end
end;

procedure DescElements;

procedure SetElement(No :integer; Na :TString; So :boolean; Mn,Rn,Rm :integer);
begin
  with Element[No] do
  begin
    Name:=Na; ReqNodes:=Rn; MinNodes:=Mn; ReqMemory:=Rm; Source:=So
  end
end;

begin
  SetElement(1,'LOW',true,1,10,0);
  SetElement(2,'HIGH',true,1,10,0);
  SetElement(3,'ASK',true,1,10,0);
  SetElement(4,'AND',false,3,11,0);
  SetElement(5,'OR',false,3,11,0);
  SetElement(6,'NOT',false,2,2,0);
  SetElement(7,'NAND',false,3,11,0);
  SetElement(8,'NOR',false,3,11,0);
  SetElement(9,'XOR',false,3,3,0);
  SetElement(10,'XNOR',false,3,3,0);
  SetElement(11,'RC',false,2,2,1);
  SetElement(12,'CNTR',true,1,16,1)
end;

procedure InitElement(No :integer);
var
  M :MemType;
begin
  case No of
    11 : Mem[Pc+2]:=0;
    12 : M[Pc+16]:=low
  end
end;

procedure SetLow;
var
  Q :integer;
begin
  for Q:=0 to ActNum-1 do ActState[Mem[Pc+Q]]:=Low
end;

procedure SetHigh;
var
  Q :integer;
begin
  for Q:=0 to ActNum-1 do ActState[Mem[Pc+Q]]:=High
end;

procedure Ask;
var
  C :char;
  Q :integer;
begin
  for Q:=0 to ActNum-1 do
  begin
    PrintI(1,35,'New level of '+Node[Mem[Pc+Q]]+':');
    repeat
      read(kbd,C); C:=upcase(C)
    until C in ['L','H','N'];
    if C <> 'N' then ActState[Mem[Pc+Q]]:=C = 'H';
    ClrLine(35)
  end
end;

procedure AndGate;
var
  Q :integer;
begin
  ActState[Mem[Pc+ActNum-1]]:=High;
  for Q:=0 to ActNum-2 do
    ActState[Mem[Pc+ActNum-1]]:=ActState[Mem[Pc+ActNum-1]] and ActState[Mem[Pc+Q]]
end;

procedure OrGate;
var
  Q :integer;
begin
  ActState[Mem[Pc+ActNum-1]]:=Low;
  for Q:=0 to ActNum-2 do
    ActState[Mem[Pc+ActNum-1]]:=ActState[Mem[Pc+ActNum-1]] or ActState[Mem[Pc+Q]]
end;

procedure NotGate;
begin
  ActState[Mem[Pc+1]]:=not ActState[Mem[Pc]]
end;

procedure NandGate;
begin
  AndGate;
  ActState[Mem[Pc+ActNum-1]]:=not ActState[Mem[Pc+ActNum-1]]
end;

procedure NorGate;
begin
  OrGate;
  ActState[Mem[Pc+ActNum-1]]:=not ActState[Mem[Pc+ActNum-1]]
end;

procedure XorGate;
begin
  ActState[Mem[Pc+2]]:=ActState[Mem[Pc]] xor ActState[Mem[Pc+1]]
end;

procedure XnorGate;
begin
  XorGate;
  ActState[Mem[Pc+2]]:=not ActState[Mem[Pc+2]]
end;

procedure Rc;
begin
  ActState[Mem[Pc+1]]:=Mem[Pc+2] = 1;
  Mem[Pc+2]:=ord(ActState[Mem[Pc]])
end;

procedure Cntr;
var
  Q :integer;
  M :MemType;
begin
  Unpack(Mem[Pc+16],M);
  for Q:=0 to ActNum-1 do ActState[Mem[Pc+Q]]:=M[Q];
  Mem[Pc+16]:=Mem[Pc+16]+1
end;

procedure Simulate;
var
  Display :array[1..MaxView] of BigString;
  Rule    :BigString;
  Q,D,L   :integer;
  C       :char;

procedure Call(Op :integer);
begin
  ActNum:=0;
  while (Mem[Pc+ActNum] <> 0) and (ActNum < Element[Op].ReqNodes) do ActNum:=ActNum+1;
  case Op of
    1  : SetLow;
    2  : SetHigh;
    3  : Ask;
    4  : AndGate;
    5  : OrGate;
    6  : NotGate;
    7  : NandGate;
    8  : NorGate;
    9  : XorGate;
    10 : XnorGate;
    11 : Rc;
    12 : Cntr
  end
end;

procedure Show;
begin
  for Q:=1 to ViewPtr do
  begin
    delete(Display[Q],1,1);
    if ActState[View[Q]] then C:=chr($7F) else C:=chr($5F);
    Display[Q]:=Display[Q]+C
  end;
  delete(Rule,1,1);
  Rule:=Rule+chr(L+47);
  for Q:=1 to ViewPtr do PrintN(11,3-D+Q*D,Display[Q]);
  PrintI(11,2,Rule); PrintI(11,33,Rule);
  L:=L mod 10+1
end;

begin
  clrscr; gotoxy(96,35);
  PrintI(1,1,'                                *** The Logic-Simulator V'+Version+' ***                                ');
  PrintN(1,2,'Node'); PrintN(1,33,'Node');
  L:=1;
  if ViewPtr < MaxView div 2 then D:=2 else D:=1;
  for Q:=1 to ViewPtr do
  begin
    PrintN(1,-D+3+Q*D,copy(Node[View[Q]]+'            ',1,10));
    Display[Q]:='______________________________________________________________________________________'
  end;
  Rule:='                                                                                      ';
  for Q:=1 to NumOfNodes do ActState[Q]:=Low;
  repeat
    Pc:=2;
    for Q:=1 to NumOfParts do
    begin
      if Element[Mem[Pc-1]].Source then Call(Mem[Pc-1]);
      Pc:=Pc+UsedMem(Mem[Pc-1])
    end;
    NewRun:=true;
    repeat
      RefState:=ActState;
      Pc:=2;
      for Q:=1 to NumOfParts do
      begin
        if not Element[Mem[Pc-1]].Source then Call(Mem[Pc-1]);
        Pc:=Pc+UsedMem(Mem[Pc-1])
      end;
      Changes:=false;
      for Q:=1 to NumOfNodes do if RefState[Q] <> ActState[Q] then Changes:=true;
      if (InStable = 'Y') and (NewRun or Changes) then Show;
      NewRun:=false
    until (not Changes) or keypressed;
    if InStable <> 'Y' then Show;
    if Changes then
    begin
      PrintI(1,35,'Simulation aborted.');
      read(kbd,C); read(kbd,C);
      ClrLine(35)
    end;
    PrintI(1,35,'Do you wish to quit ? ');
    repeat
      read(kbd,C); C:=upcase(C)
    until C in ['Y','N'];
    ClrLine(35)
  until C = 'Y';
  clrscr
end;

procedure ShowElements;
var
  Q :integer;
  C :char;
begin
  clrscr;
  writeln('Num.      Name Source MinNodes MaxNodes Memory');
  for Q:=1 to NumOfElements do
    with Element[Q] do
      writeln(Q:3,'.',Name:10,' ',Source:6,MinNodes:9,ReqNodes:9,ReqMemory:7);
  read(kbd,C)
end;

procedure SplitLine;
var
  Q,S :integer;
begin
  for Q:=1 to length(Line) do Line[Q]:=upcase(Line[Q]);
  S:=-1; NumOfPieces:=0; Line:=Line+' ';
  for Q:=1 to length(Line) do
    if (S = -1) and not (ord(Line[Q]) in [ord(Delimiter),10,13]) then S:=Q
    else
      if (S <> -1) and (Line[Q] = Delimiter) then
      begin
        NumOfPieces:=NumOfPieces+1;
        Piece[NumOfPieces]:=copy(Line,S,Q-S);
        S:=-1
      end
end;

function NodeNo(N :TString):integer;
var
  Q,T :integer;
begin
  T:=0;
  for Q:=1 to NumOfNodes do
    if Node[Q] = N then T:=Q;
  if T = 0 then
    if NumOfNodes = MaxNodes then
    begin
      writeln('Error : Too many nodes');
      Error:=true
    end
    else
    begin
      NumOfNodes:=NumOfNodes+1;
      T:=NumOfNodes;
      Node[T]:=N
    end;
  NodeNo:=T
end;

begin
  clrscr;
  DescElements;
  writeln('*** The Logic-Simulator V',Version,' ***');
  writeln;
  write('Show instable states ? '); readln(InStable);
  writeln;
  InStable:=upcase(InStable[1]);
  write('Name of map to be simulated :'); readln(MapName);
  if MapName = '?' then ShowElements
  else
  begin
    {$I-} assign(Map,MapName); reset(Map); {$I+}
    if ioresult <> 0 then
    begin
      writeln;
      writeln('Error : Cannot find map',chr(7));
      writeln
    end
    else
    begin
      Pc:=1; NumOfParts:=0; NumOfNodes:=0; ViewPtr:=0;
      Error:=false;
      while (not eof(Map)) and (not Error) do
      begin
        readln(Map,Line); SplitLine;
        if NumOfPieces <> 0 then
          if Piece[1] = 'VIEW' then
            for Q:=2 to NumOfPieces do
              if ViewPtr <= MaxView then
              begin
                ViewPtr:=ViewPtr+1;
                View[ViewPtr]:=NodeNo(Piece[Q])
              end
              else
              begin
                if not Error then writeln('Error : You cannot view so many nodes',chr(7));
                Error:=true
              end
          else
          begin
            NumOfParts:=NumOfParts+1;
            ElementNo:=-1;
            for Q:=1 to NumOfElements do
              if Element[Q].Name = Piece[1] then ElementNo:=Q;
            if ElementNo = -1 then
            begin
              writeln('Error : Unknown element :',Line,chr(7));
              Error:=true
            end
            else
            if not (NumOfPieces-1 in [Element[ElementNo].MinNodes..
               Element[ElementNo].ReqNodes]) then
            begin
              writeln('Error : Pin - Mismatch :',Line,chr(7));
              Error:=true
            end
            else
            if Pc+UsedMem(ElementNo) > MemSize then
            begin
              writeln('Error : Out of Memory',chr(7));
              Error:=true
            end
            else
            begin
              Mem[Pc]:=ElementNo;
              Q:=1;
              while (Q < NumOfPieces) and (not Error) do
              begin
                Q:=Q+1;
                Mem[Pc+Q-1]:=NodeNo(Piece[Q])
              end;
              if not Error then
              begin
                for Q:=NumOfPieces to Element[ElementNo].ReqNodes do Mem[Pc+Q]:=0;
                Pc:=Pc+1; InitElement(ElementNo);
                Pc:=Pc+UsedMem(ElementNo)-1
              end
            end
          end
      end;
      close(Map);
      if not Error then
        if ViewPtr = 0 then writeln('Error : Nothing to View',chr(7))
        else Simulate
    end
  end;
  writeln('Good bye.');
end.
