program RWCPM;

{Dateitransferprogramm zwischen CP/M-86 und MS-DOS          }
{fuer IBM-kompatible Rechner ueber Interrupt 13H            }
{CP/M-Disketten: IBM-Format 8 Sektoren, 40 Tracks, einseitig}

const
  Offset     = 1;
  BlkPerTrk  = 4;
  SecPerBlk  = 2;
  RecPerBlk  = 8;
  RecPerEint = 128;
  MaxDir     = 64;
  MaxBlk     = 155;

type
  DEintrag  = array[0..31] of byte;
  Directory = array[1..MaxDir] of DEintrag;
  Rec       = array[0..127] of byte;
  Daten     = array[1..128] of Rec;
  Filename  = string[64];
  Regpack   = record
    AX,BX,CX,DX,BP,SI,DI,DS,ES,Flags :integer
  end;

var
  DirBuf                    :Directory;
  DatBuf                    :Daten;
  Wahl                      :char;
  PhysDsk,FrDirCnt,FrBlkCnt :integer;
  BufVoll                   :boolean;
  FreiDir                   :set of 1..MaxDir;
  FreiBlk                   :set of 2..MaxBlk;

procedure Menu;
begin
  writeln;
  writeln('Moegliche Kommandos:');
  writeln;
  writeln('(1) CP/M-Directory anzeigen');
  writeln('(2) Neue CP/M-Diskette anmelden');
  writeln('(3) Datei von CP/M nach DOS kopieren');
  writeln('(4) Datei von DOS nach CP/M kopieren');
  writeln('(5) CP/M-File loeschen');
  writeln('(6) Neues physikalisches Laufwerk fuer CP/M-Diskette');
  writeln('(7) Programm Ende');
  writeln('(?) Diese Liste anzeigen')
end;

procedure Init;
begin
  BufVoll:=false;
  clrscr;
  writeln('Dateitransfer zwischen CP/M und MS-/PC-DOS');
  writeln;
  repeat
    write('Welches phys. Laufwerk enthaelt die CP/M-Diskette? (0-3) ');
    readln(PhysDsk)
  until PhysDsk in [0..3];
  Menu
end;

procedure RW_Block(Blk :integer; var Buffer; Command :integer);
var
  Track,Sektor,Retry :integer;
  Register           :Regpack;
begin
  Track:=Blk div BlkPerTrk+Offset;
  Sektor:=(Blk mod BlkPerTrk)*SecPerBlk+1;
  Retry:=0;
  repeat
    with Register do
    begin
      AX:=(Command shl 8)+2;    {2 Sektoren    }
      CX:=(Track shl 8)+Sektor;
      DX:=$0000+PhysDsk;        {Kopf und Drive}
      ES:=seg(Buffer);          {DMA-Segment   }
      BX:=ofs(Buffer)           {DMA-Offset    }
    end;
    intr($13,Register);
    if (Register.AX and $FF00) <> 0 then
    begin
      Retry:=Retry+1;
      if Retry > 3 then
      begin
        writeln('Fehler beim Diskettenzugriff CP/M-Disk');
        halt
      end
    end
  until (Register.AX and $FF00) = 0
end;

procedure ScanFrei;
var
  I,J,RecCount :integer;
begin
  FreiDir:=[]; FrDirCnt:=0;
  FreiBlk:=[2..MaxBlk]; FrBlkCnt:=MaxBlk-1;
  for I:=1 to MaxDir do
    if DirBuf[I,0] = $E5 then
    begin
      FreiDir:=FreiDir+[I];
      FrDirCnt:=FrDirCnt+1
    end
    else
    begin
      RecCount:=DirBuf[I,15];
      J:=16;
      while RecCount > 0 do
      begin
        FreiBlk:=FreiBlk-[DirBuf[I,J]];
        FrBlkCnt:=FrBlkCnt-1;
        J:=J+1;
        RecCount:=RecCount-RecPerBlk
      end
    end
end;

procedure Lese_Directory;
begin
  RW_Block(0,DirBuf[1,0],2);
  RW_Block(1,DirBuf[33,0],2);
  BufVoll:=true;
  ScanFrei
end;

procedure Zeige_Directory;
var
  I,J :integer;
begin
  if not BufVoll then Lese_Directory;
  for I:=1 to MaxDir do
    if (DirBuf[I,0] <> $E5) and (DirBuf[I,12] = 0) then
    begin
      for J:=1 to 8 do write(chr(DirBuf[I,J]));
      write(' ');
      for J:=9 to 11 do write(chr(DirBuf[I,J] and $7F));
      write('    ')
    end;
  writeln;
  writeln(FrBlkCnt,' kByte und ',FrDirCnt,' Dir-Eintraege frei')
end;

function StUpcase(Name :Filename):Filename;
var
  I    :integer;
  Ausg :Filename;
begin
  Ausg:='';
  for I:=1 to length(Name) do Ausg:=Ausg+upcase(Name[I]);
  StUpcase:=Ausg
end;

function FName(Eintrag :integer):Filename;
var
  I    :integer;
  Name :Filename;
begin
  I:=1; Name:='';
  repeat
    Name:=Name+chr(DirBuf[Eintrag,I]);
    I:=I+1
  until (I = 9) or (DirBuf[Eintrag,I] = $20);
  if (DirBuf[Eintrag,9] and $7F) <> $20 then
  begin
    Name:=Name+'.';
    I:=9;
    repeat
      Name:=Name+chr(DirBuf[Eintrag,I] and $7F);
      I:=I+1
    until (I = 12) or ((DirBuf[Eintrag,I] and $7F) = $20)
  end;
  FName:=Name
end;

function Erlaubt(var F :Filename):boolean;
const
  ZSatz :set of char = ['A'..'Z','0'..'9','!','#'..'&','+','-','/'];
var
  I,J,K :integer;
  Valid :boolean;
begin
  if length(F) > 2 then
    if F[2] = ':' then delete(F,1,2);
  Valid:=false;
  K:=length(F);
  if K > 0 then
    if F[1] <> '.' then
    begin
      I:=1;
      while (I <= 8) and (I <= K) and (F[I] in ZSatz) do I:=I+1;
      if I = K+1 then Valid:=true
      else
        if F[I] = '.' then
        begin
          J:=I+1;
          while (J <= K) and (J <= I+3) and (F[J] in ZSatz) do J:=J+1;
          Valid:=(J = K+1)
        end
    end;
  Erlaubt:=Valid
end;

function Finde_Eintrag(CfName :Filename; Extent :integer):integer;
var
  I :integer;
begin
  I:=1;
  while (I <= MaxDir) and ((DirBuf[I,0] = $E5) or (DirBuf[I,12] <> Extent)
                                                  or (FName(I) <> CfName)) do
    I:=I+1;
  Finde_Eintrag:=I
end;

procedure Kopiere_cm;
var
  CfName,MfName             :Filename;
  MFile                     :file;
  I,J,Count,Extent,RecCount :integer;
  Error                     :boolean;
begin
  if not BufVoll then Lese_Directory;
  write('Bitte CP/M-Filenamen eingeben: ');
  readln(CfName);
  CfName:=StUpcase(CfName);
  if Finde_Eintrag(CfName,0) <= MaxDir then
  begin
    repeat
      write('Bitte DOS-Filenamen eingeben: ');
      readln(MfName);
      assign(MFile,MfName);
      {$I-} rewrite(MFile); {$I+}
      Error:=ioresult <> 0;
      if Error then writeln('DOS-Datei kann nicht geschrieben werden!')
    until not Error;
    Extent:=0;
    repeat
      I:=Finde_Eintrag(CfName,Extent);
      RecCount:=DirBuf[I,15];
      Count:=1;
      J:=16;
      while RecCount > 0 do
      begin
        RW_Block(DirBuf[I,J],DatBuf[Count,0],2);
        J:=J+1;
        Count:=Count+RecPerBlk;
        RecCount:=RecCount-RecPerBlk
      end;
      blockwrite(MFile,DatBuf,DirBuf[I,15]);
      Extent:=Extent+1
    until DirBuf[I,15] < RecPerEint;
    close(MFile)
  end
  else writeln('File nicht im Directory!')
end;

procedure NamensEintrag(FName :Filename; Extent,Eintrag :integer);
var
  I,J    :integer;
  Fehler :boolean;
begin
  DirBuf[Eintrag,0]:=0;
  I:=1;
  while (I <= length(FName)) and (FName[I] <> '.') do
  begin
    DirBuf[Eintrag,I]:=ord(FName[I]);
    I:=I+1
  end;
  J:=I;
  while J <= 8 do
  begin
    DirBuf[Eintrag,J]:=$20;
    J:=J+1
  end;
  if FName[I] = '.' then I:=I+1;
  while (J <= 11) and (I <= length(FName)) do
  begin
    DirBuf[Eintrag,J]:=ord(FName[I]);
    I:=I+1; J:=J+1
  end;
  while J <= 11 do
  begin
    DirBuf[Eintrag,J]:=$20;
    J:=J+1
  end;
  DirBuf[Eintrag,12]:=Extent;
  for I:=13 to 31 do DirBuf[Eintrag,I]:=0
end;

procedure Kopiere_mc;
var
  CfName,MfName :Filename;
  MFile         :file;
  RecCount,Extent,RecStoRead,DatZeiger,EintZeiger,DirZeiger,I,J :integer;
begin
  if not BufVoll then Lese_Directory;
  write('Bitte DOS-Filenamen eingeben: ');
  readln(MfName);
  assign(MFile,MfName);
  {$I-} reset(MFile); {$I+}
  if ioresult = 0 then
  begin
    repeat
      write('Bitte CP/M-Filenamen eingeben: ');
      readln(CfName);
      CfName:=StUpcase(CfName)
    until Erlaubt(CfName);
    RecCount:=filesize(MFile);
    if ((RecCount-1) div 8 < FrBlkCnt)
                     and (RecCount div RecPerEint < FrDirCnt) then
    begin
      I:=0;
      while Finde_Eintrag(CfName,I) <= MaxDir do
      begin
        J:=Finde_Eintrag(CfName,I);
        DirBuf[J,0]:=$E5;
        I:=I+1
      end;
      ScanFrei;
      Extent:=0;
      DirZeiger:=1; J:=1;
      while RecCount >= 0 do
      begin
        while not (DirZeiger in FreiDir) do DirZeiger:=DirZeiger+1;
        FreiDir:=FreiDir-[DirZeiger];
        FrDirCnt:=FrDirCnt-1;
        NamensEintrag(CfName,Extent,DirZeiger);
        if RecCount > RecPerEint then RecStoRead:=RecPerEint
        else RecStoRead:=RecCount;
        DirBuf[DirZeiger,15]:=RecStoRead;
        blockread(MFile,DatBuf,RecStoRead);
        DatZeiger:=1; EintZeiger:=16;
        while RecStoRead > 0 do
        begin
          while not (J in FreiBlk) do J:=J+1;
          RW_Block(J,DatBuf[DatZeiger,0],3);
          DirBuf[DirZeiger,EintZeiger]:=J;
          FreiBlk:=FreiBlk-[J];
          FrBlkCnt:=FrBlkCnt-1;
          J:=J+1;
          DatZeiger:=DatZeiger+RecPerBlk;
          EintZeiger:=EintZeiger+1;
          RecStoRead:=RecStoRead-RecPerBlk
        end;
        Extent:=Extent+1;
        RecCount:=RecCount-RecPerEint
      end;
      RW_Block(0,DirBuf[1,0],3);
      RW_Block(1,DirBuf[33,0],3)
    end
    else writeln('Diskette oder Directory voll!')
  end
  else writeln('DOS-File nicht gefunden!')
end;

procedure Loesche;
var
  CfName :Filename;
  I,J    :integer;
begin
  write('Welches CP/M-File soll geloescht werden ? ');
  readln(CfName);
  if length(CfName) > 0 then
  begin
    CfName:=StUpcase(CfName);
    J:=Finde_Eintrag(CfName,0);
    if J <= MaxDir then
    begin
      DirBuf[J,0]:=$E5;
      I:=1;
      while Finde_Eintrag(CfName,I) <= MaxDir do
      begin
        J:=Finde_Eintrag(CfName,I);
        DirBuf[J,0]:=$E5;
        I:=I+1
      end;
      RW_Block(0,DirBuf[1,0],3);
      RW_Block(1,DirBuf[33,0],3);
      ScanFrei
    end
    else writeln('File nicht im Directory!')
  end
end;

begin {RWCPM}
  Init;
  writeln;
  repeat
    repeat
      write('Kommando: ');
      read(Wahl);
      writeln;
      if not (Wahl in ['1'..'7','?']) then writeln('????')
    until Wahl in ['1'..'7','?'];
    case Wahl of
      '1' : Zeige_Directory;
      '2' : begin
              Lese_Directory;
              Zeige_Directory
            end;
      '3' : Kopiere_cm;
      '4' : Kopiere_mc;
      '5' : Loesche;
      '6' : Init;
      '?' : Menu
    end
  until Wahl = '7'
end.
